"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const DataLoader = require("dataloader");
const invariant = require("invariant");
const bluestream_1 = require("bluestream");
const scanCommands = ['scan', 'sscan', 'hscan', 'zscan', 'scanBuffer', 'sscanBuffer', 'hscanBuffer', 'zscanBuffer'];
const pubSubCommands = ['subscribe', 'unsubscribe', 'publish', 'psubscribe', 'punsubscribe'];
class RedisLoader {
    constructor({ redis, logger = () => { } } = {}) {
        invariant(redis, '"redis" is required');
        this.resetStats();
        this._redis = redis;
        this._dataLoader = new DataLoader((commands) => __awaiter(this, void 0, void 0, function* () {
            const start = Date.now();
            this._stats.commands = commands;
            this._stats.sentCommandCount = commands.length;
            this._stats.sentCommandCountTotal += commands.length;
            const setEndStats = (error, results) => {
                const end = Date.now();
                const timeInRedis = end - start;
                this._stats.tripCountTotal++;
                this._stats.timeInRedis = timeInRedis;
                this._stats.timeInRedisTotal += timeInRedis;
                if (error) {
                    this._stats.responseCount = 1;
                    this._stats.responseCountTotal += 1;
                }
                else {
                    this._stats.responseCount = results.length;
                    this._stats.responseCountTotal += results.length;
                }
            };
            try {
                const results = yield redis.multi(commands).exec();
                setEndStats(null, results);
                const parsedResults = results.map(([err, data]) => {
                    if (err) {
                        logger(err, this.stats);
                        throw err;
                    }
                    return data;
                });
                logger(null, this.stats);
                return parsedResults;
            }
            catch (err) {
                setEndStats(err);
                if (Array.isArray(err.previousErrors)) {
                    err.message = `${err.message} ${err.previousErrors.map(e => e && e.message)}`;
                }
                logger(err, this.stats);
                throw err;
            }
        }), { cache: false });
        redis.getBuiltinCommands().forEach(command => {
            this[command] = (...args) => this._dataLoader.load([command, ...args]);
            const bufferCmd = `${command}Buffer`;
            if (redis[bufferCmd]) {
                this[bufferCmd] = (...args) => this._dataLoader.load([bufferCmd, ...args]);
            }
        });
        scanCommands.forEach(command => {
            this[`${command}Stream`] = (key, options) => {
                if (command === 'scan' || command === 'scanBuffer') {
                    options = key;
                    key = null;
                }
                return new ScanStream(Object.assign({ key, redis: this, command }, options));
            };
        });
        pubSubCommands.forEach(command => {
            this[command] = (...args) => redis[command](...args);
            this[`${command}Buffer`] = (...args) => redis[`${command}Buffer`](...args);
        });
        Object.keys(events_1.EventEmitter.prototype).forEach(key => {
            if (typeof events_1.EventEmitter.prototype[key] === 'function') {
                this[key] = (...args) => redis[key](...args);
            }
            else {
                Object.defineProperty(this, key, { value: redis[key], writable: false });
            }
        });
    }
    get stats() {
        return Object.assign({}, this._stats);
    }
    resetStats({ tripCountTotal = 0, commands = null, sentCommandCount = 0, sentCommandCountTotal = 0, responseCount = 0, responseCountTotal = 0, timeInRedis = 0, timeInRedisTotal = 0 } = {}) {
        this._stats = {
            tripCountTotal,
            commands,
            sentCommandCount,
            responseCount,
            sentCommandCountTotal,
            responseCountTotal,
            timeInRedis,
            timeInRedisTotal
        };
    }
}
exports.default = RedisLoader;
class ScanStream extends bluestream_1.ReadStream {
    constructor(opts = {}) {
        invariant(opts.redis, '"opts.redis" is required');
        super(opts);
        this._redis = opts.redis;
        this._command = opts.command;
        this._opts = opts;
        this._nextCursor = '0';
    }
    _read() {
        return __awaiter(this, void 0, void 0, function* () {
            const { _opts } = this;
            const args = [this._nextCursor];
            if (_opts.key) {
                args.unshift(_opts.key);
            }
            if (_opts.match) {
                args.push('MATCH', _opts.match);
            }
            if (_opts.count) {
                args.push('COUNT', _opts.count);
            }
            const [nextCursor, redisIds] = yield this._redis[this._command](args);
            this._nextCursor = nextCursor instanceof Buffer ? nextCursor.toString() : nextCursor;
            this.push(redisIds);
            if (this._nextCursor === '0') {
                this.push(null);
            }
        });
    }
}
